import os
import shutil
import json
import goenrich
import pandas as pd
import numpy as np

import phos.algo.anat

def inference(exp, scores, network_undirected, anat, go, task_id, db, **kwargs):
    active = scores[scores['Significant']]['GeneID']
    # LOCAL
    # subnetwork = phos.algo.anat.local_network(network_undirected,
    #    active, exp, task_id=task_id, **anat)
    subnetwork = phos.algo.anat.remote_network(task_id, network_undirected,
            active, anat.get('anchor', None)).astype(int)
    if subnetwork is None:
        print("Subnetwork did not contain any edges")
        subnetwork = pd.DataFrame({'s':[-1], 't':['-1']})
    # GO enrichment
    O = goenrich.obo.ontology(db['go']['ontology'])
    annotations = goenrich.read.gene2go(db['go']['annotations'])
    nodes = pd.DataFrame({'GeneID' :
        list(set(network_undirected['kin']) | set(network_undirected['sub']))})
    background = goenrich.tools.generate_background(annotations, nodes, 'GO_ID', 'GeneID')
    query = (set(subnetwork['s']) | set(subnetwork['t'])) - set([anat.get('anchor', None)])
    goenrich.enrich.propagate(O, background, 'scores')
    go_scores = goenrich.enrich.analyze(O, query, 'scores', **go)
    return subnetwork, go_scores

import networkx as nx
from networkx.readwrite import json_graph

def create_graph(exp, scores, network, task_id, db, anchor=None):
    """ generate result graph
    :param exp: pandas.DataFrame with columns: Symbol, GeneID, Amino.Acid, Position, avg
    :param scores: protein activity scores 
    :param network: pandas.DataFrame with columns 's' -> 't'
    :param anchor: anchor id
    """
    terminals = set(scores[scores['Significant']]['GeneID'])
    G = nx.from_edgelist([[int(x) for x in y] for y in network[['s','t']].values])
    df = (exp[['Symbol', 'GeneID', 'Amino.Acid', 'Position', 'avg']]
            [exp['GeneID'].isin(network['s']) | exp['GeneID'].isin(network['t'])])
    node_attributes = (df.groupby('GeneID')
            .apply(lambda x : [{'AA': a, 'POS': int(p), 'AVG': v, 'NUM': len(x)} for a,p,v in
                zip(x['Amino.Acid'].values, x['Position'].values, x['avg'].values)]))
    geneinfo_table = pd.read_table(db['geneinfo'], comment='#', usecols=[1, 2], header=None)
    gene_name = dict(geneinfo_table.dropna().values)
    for n in G:
        G.node[n]['residues'] = node_attributes.get(n, [{'NUM' : 0}])
        G.node[n]['type'] = 'terminal' if n in terminals else 'connector'
        G.node[n]['name'] = gene_name.get(n, n)
    if anchor in G and anchor is not None:
        G.node[anchor]['type'] = 'anchor'
    return G

def draw(exp, scores, network, task_id, template_dir, db, anchor=None):
    """ generate interactive result graph based on d3js
    
    check `phos/algo/anat.html` for the javascript part of the visualization

    :param exp: pandas.DataFrame with columns: Symbol, GeneID, Amino.Acid, Position, avg
    :param scores: protein activity scores 
    :param network: pandas.DataFrame with columns 's' -> 't'
    :param anchor: anchor id
    :returns HTML: the generated graph
    """
    G = create_graph(exp, scores, network, task_id, db, anchor)
    graph = json_graph.node_link_data(G)
    node_index = {n['id']: i for i,n in enumerate(graph['nodes'])}
    graph['links'] = [{'source': node_index[link['source']], 'target': node_index[link['target']]} for link in graph['links']]
    from jinja2 import Environment, FileSystemLoader
    env = Environment(loader=FileSystemLoader(template_dir))
    HTML = env.get_template('result.html').render(task_id=task_id, graph=json.dumps(graph))
    return HTML

def _rnd_go(exp, _score, network_undirected, anat, go, seed):
    np.random.seed(seed)
    rnd = (_score[['Significant']]
            .reindex(np.random.permutation(_score.index))
            .reset_index(drop=True))
    rnd['GeneID'] = _score['GeneID']
    _anat = anat.copy()
    rand_dir = os.path.join(anat['anat_dir'], str(seed))
    _anat.update({ 'anat_dir' : rand_dir,
        'outfile' : os.path.join(anat['anat_dir'], 'random')})
    _, go_rand = inference(exp, rnd, network_undirected, _anat, go)
    shutil.rmtree(rand_dir)
    return go_rand.dropna()

def inference_empiric_pvalue(exp, scores, network_undirected, anat, go, **kwargs):
    from joblib import Parallel, delayed
    perm = anat['permutations']
    subnetwork, go_scores = inference(exp, scores, network_undirected, anat, go)
    _anat = anat.copy()
    if 'viz_out' in _anat:
        del _anat['viz_out']
    MAX_INT = np.iinfo(np.int32).max
    seeds = np.random.randint(MAX_INT, size=perm)
    _scores = scores.reset_index(drop=True)
    rands = Parallel(n_jobs=35)(delayed(_rnd_go)(exp, _scores, network_undirected, _anat, go, seed) for seed in seeds)
    random = pd.concat(rands)[['term', 'p']].rename(columns={'p' : 'p_random'})
    df = (pd.merge(go_scores, random, how='left')
            .dropna(subset=['q'])
            .fillna(0))
    df['p>p_random'] = df['p'] > df['p_random']
    p_empiric = ((df.groupby('term')['p>p_random'].sum() / perm)
            .to_frame(name='p_empiric')
            .reset_index())
    result = pd.merge(go_scores, p_empiric)
    return subnetwork, result



from phos.util import read_json
from os.path import dirname, join, isfile

descriptions = {
        'go' : 'Gene Ontology enrichment',
        'max_category_size' : 'Maximal category size',
        'min_category_size' : 'Minimal category size',
        'max_category_depth' : 'Maximal ontolgy-tree depth',
        'anat' : 'Network reconstruction',
        'anchor' : 'Signaling source (anchor)',
        'alpha' : 'Global-local optimization tradeoff ($\\alpha$)',
        'activity' : 'Signaling functionality scoring',
        'permutations' : 'Permutations ($r$)',
        'min_size' : 'Required number of observations ($n_0$)',
        'ppi-network' : 'Protein-protein interaction network',
        'degree_threshold' : 'Maximal protein degree',
        'confidence' : 'Edge-confidence threshold',
        'side' : 'Test sidedness (greater, twosided, lesser)'
        }

def make_defaults(root_dir, work='work', example='example', db='db'):
    work_dir = join(root_dir, work)
    example_dir = join(root_dir, example)
    db_dir = join(root_dir, db)
    steinprt = join(root_dir, 'bin', 'steinprt')
    db = { "geneinfo" : join(db_dir, 'geneinfo', 'Homo_sapiens.gene_info.gz'),
            "uniprot_mapping" : join(db_dir, 'uniprot',
                'HUMAN_9606_idmapping.dat.gz'),
            'uniprot_humsavar' : join(db_dir, 'uniprot', 'humsavar.txt'),
            'ppi-network' : join(db_dir, 'anat', 'H_sapiens.net'),
            'go' : {
                'ontology' : join(db_dir, 'go', 'go-basic.obo'),
                'annotations' : join(db_dir, 'go', 'gene2go.gz')
                },
            'phosphosite' : {
                'all_sites' : join(db_dir, 'phosphosite',
                    'Phosphorylation_site_dataset.gz'),
                'disease_associated_sites' : join(db_dir, 'phosphosite',
                    'Disease-associated_sites.csv'),
                'regulatory_sites' : join(db_dir, 'phosphosite',
                    'Regulatory_sites.csv'),
                'kinase_substrate_interactions' : join(db_dir, 'phosphosite',
                    'Kinase_Substrate_Dataset.gz')
                }
            }
    return {'root': root_dir, 'work': work_dir, 'example': example_dir, 'db': db, 'steinprt': steinprt}



#!/usr/bin/env python
import os
from os.path import join, abspath, dirname
import io
import zipfile
from glob import glob
from flask import Flask, url_for, redirect, render_template, request, jsonify, send_from_directory, send_file
from phos.defaults import make_defaults, descriptions
import phos.util
defaults = make_defaults(abspath(dirname(dirname(__file__))))
template_dir = abspath(join(defaults['root'], 'templates'))
app = Flask(__name__,
        template_folder=template_dir,
        static_folder=abspath(join(defaults['root'], 'static')))
app.config['CELERY_BROKER_URL'] = 'redis://localhost:6379/0'
app.config['CELERY_RESULT_BACKEND'] = 'redis://localhost:6379/0'
app.config['CELERY_ACCEPT_CONTENT'] = ['json']
app.config['CELERY_TASK_SERIALIZER'] = 'json'
app.config['CELERY_RESULT_SERIALIZER'] = 'json'

from celery import Celery
from celery.utils import uuid
celery = Celery(app.name, broker=app.config['CELERY_BROKER_URL'])
celery.conf.update(app.config)

import phos.pipeline

import json
import redis
def task_key(task_id):
    return 'task:{}:progress'.format(task_id)

redis_db = redis.StrictRedis()
def make_set_progress(task_id):
    def set_progress(message):
        redis_db.set(task_key(task_id), json.dumps(message))
    return set_progress

def get_progress(task_id):
    raw_message = redis_db.get(task_key(task_id))
    if not raw_message:
        return "no message found"
    message = json.loads(raw_message.decode('utf-8'))
    return message

@celery.task()
def phos_task(task_id, data, _parameters):
    parameters = phos.util.read_json(abspath(join(defaults['root'], 'parameters.json')))
    phos.util.deep_update(parameters, _parameters)
    return phos.pipeline.run(task_id, data, parameters, work_dir=defaults['work'], db=defaults['db'], template_dir=template_dir, set_progress = make_set_progress(task_id))

@app.route('/', methods=['GET', 'POST'])
def index():
    parameters = phos.util.read_json(abspath(join(defaults['root'], 'parameters.json')))
    parameters['anat']['anchor'] = None
    return render_template('index.html', parameters=parameters,
            descriptions=descriptions)

@app.route('/submit', methods=['POST'])
def submit():
    task_id = uuid()
    TASK_DIR = os.path.join(defaults['work'], task_id)
    os.makedirs(TASK_DIR)
    file = request.files['data']
    filename = os.path.join(TASK_DIR, 'data.tab')
    file.save(filename)
    form = request.form
    parameters = phos.util.read_json(abspath(join(defaults['root'], 'parameters.json')))
    _parameters = {k:{} for k in parameters.keys()}
    for _k,v in form.items():
        k, kk = _k.split(':')
        _parameters[k][kk] = phos.util.num(v)
    phos_task.apply_async(args=(task_id, filename, _parameters), task_id=task_id)
    return render_template('submit.html', task_id=task_id)

@app.route('/get/<task_id>')
def show_result(task_id):
    return send_from_directory(os.path.join(defaults['work'], task_id),
            filename='result.html')

@app.route('/download/<task_id>')
def download_result(task_id):
    memory_zip = io.BytesIO()
    with zipfile.ZipFile(memory_zip, 'w') as zf:
        for path in glob(os.path.join(defaults['work'], task_id, '*')):
            rpath = os.path.relpath(path)
            arcname = os.path.join(*os.path.split(rpath)[1:])
            zf.write(rpath, arcname=arcname)
    memory_zip.seek(0)
    return send_file(memory_zip, attachment_filename='{}.zip'.format(task_id),
            as_attachment=True)

@app.route('/status/<task_id>')
def status(task_id):
    task = phos_task.AsyncResult(task_id)
    info = str(task.info) if task.state == 'FAILURE' else ''
    message = get_progress(task_id)
    print('status info', info)
    return jsonify({'state' : task.state, 'info': info, 'message' : message})

if __name__ == '__main__':
    import argparse

    parser = argparse.ArgumentParser(description='PHOTON server')
    parser.add_argument('--debug', action='store_const',
            const=True, default=False, help='start server in debug mode. \
                    Do not set this on a production system. \
                    Allows for the execution of arbitrary code on your system')
    parser.add_argument('--host', type=str, default='0.0.0.0',
            help="server host. Change to '127.0.0.1' to be visible only on your computer")
    args = parser.parse_args()
    app.run(debug=args.debug, host=args.host)



""" utility function for SILAC (log-fold) based datasets """
import pandas as pd
import numpy as np
from scipy.stats import ttest_1samp
import phos.util.mapping as mapping
from phos.util import fdrcorrection

def threshold(data, fold_changes, t=2, **kwargs):
    return pd.DataFrame(
            (data[fold_changes] > t).all(1) | (data[fold_changes] < -t).all(1),
            columns=['rej'])

def one_sample_ttest(grouped, data, fold_changes, **kwargs):
    """ one sample ttest for differential phosphorylation of a gene,
    fdr correction at alpha 0.05 """
    data_mean = np.mean(data[fold_changes].values)
    test_results = (grouped
            .apply(lambda df: ttest_1samp(np.hstack(df[fold_changes].values), data_mean)))
    test_T = test_results.str.get(0)
    test_p = test_results.str.get(1)
    rej, q = fdrcorrection(test_p.values)
    df = pd.concat([test_T, test_p], axis=1)
    df.columns = ['T', 'p']
    df['rej'] = rej
    df['q'] = q
    return df.reset_index()

def experiment(df, fold_changes, uniprot_mapping, geneinfo, **kwargs):
    """ use 'experiments' entry to map df to GeneID and extract relevant columns
   
    >>> mapped = silac.experiment(df, **params)

    :param df: pandas.DataFrame to extract from
    :param fold_changes: list of data columns
    :param uniprot_mapping: string specifying the uniprot mapping file location
    :param geneinfo: string specifying the geninfo file location
    :returns mapped: pandas.DataFrame with the experiment mapped to geneids
    """
    site_cols = ['Amino.Acid', 'Position']
    df = (df.dropna(subset=site_cols)
            .dropna(subset=fold_changes, how='all'))
    df = df.rename(columns={fc : 'fc' for fc in fold_changes})
    geneid2symbol = pd.read_table(geneinfo, compression='gzip', comment='#',
            usecols = [1, 2], names=['GeneID', 'Symbol'])
    mapped = (mapping
            .map_protein_groups(df, 'Uniprot', 'GeneID', uniprot_mapping)
            [['GeneID'] + site_cols + ['fc']]
            .dropna()
            .drop_duplicates(subset = ['GeneID'] + site_cols)
            .merge(geneid2symbol))
    mapped['Position'] = mapped['Position'].astype(int)
    mapped['fc'] = (mapped['fc'] - mapped['fc'].mean()) / mapped['fc'].std()
    if type(mapped['fc']) is pd.Series:
        mapped['avg'] = mapped['fc']
    else: # type is pd.DataFrame
        mapped['avg'] = mapped['fc'].mean(axis=1)
    return mapped



#!/usr/bin/env python
import os
from os.path import join
import pandas as pd
import numpy as np
import json

import phos.util

import phos.data.read as reader
import phos.data.network as networker

import phos.algo.activity as activity
import phos.algo.subnetwork as subnetwork
import phos.algo.functional_sites as functional_sites

from phos.defaults import make_defaults

def print_progress(message):
    print('progress:', message)

def _run(task_id, data, parameters, db, set_progress=print_progress):
    exp = pd.read_csv(data)
    columns = ['GeneID', 'Amino.Acid', 'Position', 'avg', 'Symbol']
    if len(exp.columns) < len(columns):
	    raise ValueError('Found {} columns but expected {}. Please check that the input format is correct!'.format(len(exp.columns), len(columns)))
    if not (exp.columns == columns).all():
        raise ValueError('Column names are not {}'.format(columns))
    if not exp['GeneID'].dtypes == np.int:
        raise ValueError('GeneID could not be parsed as integer, make sure the column does not contain "NaN" or ";"')
    set_progress('1/5 read data file, constructing network...')
    network_undirected = networker.load(db=db, **parameters['ppi-network'])
    network = networker.to_directed(network_undirected)
    set_progress('2/5 constructed the network, calculating functionality scores...')
    scores = activity.empiric(exp, network, **parameters['activity'])
    if len(scores[scores['Significant']]['GeneID']) < 1:
        raise ValueError('No significant signaling functionality scores found, possibly due to insufficient data.')
    set_progress('3/5 calculated scores, infering subnetwork using ANAT...')
    subnet, go_scores = subnetwork.inference(exp, scores,
            network_undirected, db=db, task_id=task_id, **parameters)
    set_progress('4/5 inferred subnetwork, making functional site predictions')
    predictions = functional_sites.predict_functional_sites(exp, scores, db=db,
            **parameters)
    set_progress('5/5 all done')
    return exp, scores, subnet, go_scores, predictions

def run(task_id, data, parameters, template_dir, work_dir, db, set_progress=print_progress):
    if parameters['anat'].get('anchor', None) < 1:
        del parameters['anat']['anchor']
    exp, scores, subnet, go_scores, predictions = _run(task_id, data, parameters, db=db, set_progress=set_progress)
    os.makedirs(join(work_dir, task_id), exist_ok=True)
    with open(join(work_dir, task_id, 'parameters.json'), 'w') as f:
        json.dump(parameters, f, indent=1)
    scores.to_csv(join(work_dir, task_id, 'scores.csv'), index=False)
    subnet.to_csv(join(work_dir, task_id, 'subnet.csv'), index=False)
    go_scores.to_csv(join(work_dir, task_id, 'go_scores.csv'), index=False)
    HTML = subnetwork.draw(exp, scores, subnet, task_id, template_dir=template_dir,
            anchor=parameters['anat'].get('anchor', None), db=db)
    with open(join(work_dir, task_id, 'result.html'), 'w') as f:
        f.write(HTML)
    predictions[0].to_csv(join(work_dir, task_id, 'predictions.csv'), index=False)

if __name__ == '__main__':
    import argparse

    parser = argparse.ArgumentParser(description='PHOTON: PHOsphoproteomic dissecTiOn using Networks')
    parser.add_argument('data',
            help='data file')
    parser.add_argument('task_id', type=str, default=phos.util.uuid(),
            help='determines the save location, will be generated')
    parser.add_argument('parameters', help='read parameters from json')
    parser.add_argument('-r', '--root-dir', default=os.getcwd(), help='root directory')
    parser.add_argument('-t', '--template-dir', default=os.path.join(os.getcwd(), 'templates'), help='templates directory')
    parser.add_argument('-a', '--anchor', type=int, default=-1,
            help='GeneID for the subnetwork reconstruction anchor')
    parser.add_argument('-p', '--parameter', action='append',
            default = [],
            help='change parameter --parameter alpha=0.1, \
                    can be supplied mutliple times')
    args = parser.parse_args()
    defaults = make_defaults(args.root_dir)
    parameters = phos.util.read_json(args.parameters)
    parameters['anat']['anchor'] = args.anchor if args.anchor > 0 else None
    # update default values
    for _param in args.parameter:
        k,_v = _param.split('=')
        v = phos.util.num(_v)
        for cat in parameters:
            if k in parameters[cat]:
                parameters[cat][k] = v
                print(cat, k, v)
                break
    task_id = args.task_id
    os.makedirs(os.path.join(defaults['work'], task_id), exist_ok=True)
    print('running', task_id)
    run(task_id, args.data, parameters, work_dir=defaults['work'], template_dir=args.template_dir, db=defaults['db'])



import os
import subprocess

import pandas as pd
import networkx as nx
from networkx.readwrite import json_graph
import json

def run_locally(network_undirected, terminals, data, anchor,
        alpha, task_id, work_dir, viz_out=None, steinprt='steinprt', **kwargs):
    """
    run anat subnetwork inference ** requires steinprt binary **
    
    :param network_undirected: network
    :param terminals:
    :param dataset: pandas.DataFrame with columns: Symbol, GeneID, Amino.Acid, Position, avg
    :param anchor: anchor
    :param alpha: anat parameter, tradeoff between shortest path and steiner tree
    :param viz_out: file path. Result will be saved at `viz_out`.js and `viz_out`.html
    :returns subnetwork: the resulting subnetwork as a table
    """
    anat_dir = os.path.join(work_dir, task_id, 'anat')
    os.makedirs(anat_dir, exist_ok=True)
    _terminals = terminals.to_frame(name='terminal').reset_index(drop=True)
    _terminals['anchor'] = anchor
    _terminals = _terminals[_terminals['terminal'] != anchor][['anchor', 'terminal']]
    (network_undirected.iloc[:,:4]
        .to_csv(os.path.join(anat_dir, 'network'), index=False, sep='\t', header=None))
    _terminals.to_csv(os.path.join(anat_dir, 'terminals'), index=False, sep='\t', header=None)
    outfile = os.path.join(anat_dir, 'anat.out')
    output = subprocess.check_output([steinprt,
        '-n', 'network',
        '-s', 'terminals',
        '-b', '{:0.2f}'.format(alpha),
        '-f', anat_dir,
        '-r', outfile]) # -f flag ignored by steinprt
    with open(outfile + '.log', 'wb') as f:
        f.write(output)
    subnetwork = pd.read_table(outfile, sep=' ',
            names=['s', 't', 'confidence', 'transformed', 'directed'])
    return subnetwork


"""
                  CALLING ANAT SERVER FROM PYTHON EXAMPLE

    import uuid
    sessionId = str(uuid.uuid4())
    anchor = 1950
    terminals = [1120, 112, 1956]

    from anat import network
    nodes, edges = network(sessionId, anchor, terminals)

"""
import requests
import time
import xml.etree.ElementTree as ET

def get_result(sessionID):
    headers = {'Accept' : 'text/xml, multipart/related',
            'Content-Type' : 'text/xml; charset=utf-8',
            'SOAPAction' : '"getResult"'}
    data = """<?xml version='1.0' encoding='UTF-8'?>
    <S:Envelope xmlns:S="http://schemas.xmlsoap.org/soap/envelope/">
        <S:Body><ns2:sessionId xmlns:ns2="network">{}</ns2:sessionId></S:Body>
    </S:Envelope>""".format(sessionID)
    result_response = requests.post('http://anat.cs.tau.ac.il/AnatWeb/AnatServer', headers=headers, data=data)
    if not result_response.ok:
        raise ValueError('Response status not OK',
                '\nRequest', result_response.request.body,
                '\nResponse', result_response.content)
    return result_response.text

def add_edges(network_undirected):
    edgesData = ["""       <ns2:edgesData>
         <ns2:action>SET_UNDIRECTED</ns2:action>
         <ns2:additionalInfo>edge</ns2:additionalInfo>
         <ns2:confidence>{confidence}</ns2:confidence>
         <ns2:fromNodeId>{source}</ns2:fromNodeId>
         <ns2:toNodeId>{target}</ns2:toNodeId>
       </ns2:edgesData>""".format(source=source,
           target=target, confidence=confidence)
       for (source, target), confidence in
       zip(network_undirected[['kin', 'sub']].values,
           network_undirected['confidence'])]
    return '\n'.join(edgesData)
    
def make_terminals(nodes):
    return '\n'.join([
        '      <ns2:terminals>{}</ns2:terminals>'.format(node)
        for node in nodes])

def make_set(nodes):
    return '\n'.join([
        '      <ns2:set>{}</ns2:set>'.format(node) for node in nodes])

def submit_job_no_anchor(sessionId, network_undirected, terminals):
    edgesData = add_edges(network_undirected)
    setData = make_set(terminals)
    headers = { 'SOAPAction' : '"calculateProjectionSubNetwork"',
            'charset' : 'utf-8',
            'content-type' : 'text/xml' }
    data = """<?xml version="1.0" encoding="UTF-8"?>
    <S:Envelope xmlns:S="http://schemas.xmlsoap.org/soap/envelope/">
      <S:Body>
        <ns2:projectionParams xmlns:ns2="network">
          <ns2:algorithmType>PROJECTIONANALYSIS</ns2:algorithmType>
          <ns2:backGroundNetwork>
            <ns2:networkName>my_network</ns2:networkName>
            {edgesData}
          </ns2:backGroundNetwork>
          <ns2:baseNetworkFileName>E_empty.net</ns2:baseNetworkFileName>
          <ns2:edgeConstraintMap/>
          <ns2:nodeConstraintMap/>
          <ns2:sessionId>{sessionId}</ns2:sessionId>
          <ns2:title>mytitle</ns2:title>
          <ns2:lengthPenalty>25</ns2:lengthPenalty>
          <ns2:margin>0</ns2:margin>
          <ns2:algorithmType>PROJECTIONANALYSIS</ns2:algorithmType>
          <ns2:granularity>100</ns2:granularity>
          {setData}
          <ns2:subAlgorithm>CLUSTERING</ns2:subAlgorithm>
        </ns2:projectionParams>
      </S:Body>
    </S:Envelope>
    """.format(edgesData=edgesData, setData=setData, sessionId=sessionId)
    submit_response = requests.post(
            'http://anat.cs.tau.ac.il/AnatWeb/AnatServer',
            headers=headers, data=data)
    if not submit_response.ok:
        raise ValueError('Response status not OK',
                '\nRequest', submit_response.request.body,
                '\nResponse', submit_response.content)
    return submit_response.text 

def submit_job(sessionId, network_undirected, anchor, terminals):
    edgesData = add_edges(network_undirected)
    terminalsData = make_terminals(terminals)
    headers = { 'SOAPAction' : '"calculateExplanatorySubNetwork"',
            'charset' : 'utf-8',
            'content-type' : 'text/xml' }
    data = """<?xml version='1.0' encoding='UTF-8'?>
      <S:Envelope xmlns:S="http://schemas.xmlsoap.org/soap/envelope/">
        <S:Body>
        <ns2:explanatoryParameters xmlns:ns2="network">
          <ns2:algorithmType>EXPLANATORYPATHWAYS</ns2:algorithmType>
          <ns2:backGroundNetwork>
            <ns2:networkName>my_network</ns2:networkName>
            {edgesData}
          </ns2:backGroundNetwork>
          <ns2:baseNetworkFileName>E_empty.net</ns2:baseNetworkFileName>
          <ns2:edgeConstraintMap/>
          <ns2:nodeConstraintMap/>
          <ns2:sessionId>{sessionId}</ns2:sessionId>
          <ns2:title>my_title</ns2:title>
          <ns2:lengthPenalty>25</ns2:lengthPenalty>
          <ns2:margin>0</ns2:margin>
          <ns2:algorithmType>EXPLANATORYPATHWAYS</ns2:algorithmType>
          <ns2:alpha>0.25</ns2:alpha>
          <ns2:anchors>{anchor}</ns2:anchors>
          <ns2:predictTF>false</ns2:predictTF>
          <ns2:subAlgorithm>APPROXIMATE</ns2:subAlgorithm>
          {terminalsData}
          <ns2:terminalsToAnchors>false</ns2:terminalsToAnchors>
        </ns2:explanatoryParameters>
        </S:Body>
      </S:Envelope>
    """.format(edgesData=edgesData, sessionId=sessionId,
            anchor=anchor, terminalsData=terminalsData)
    submit_response = requests.post(
            'http://anat.cs.tau.ac.il/AnatWeb/AnatServer',
            headers=headers, data=data)
    if not submit_response.ok:
        raise ValueError('Response status not OK',
                '\nRequest', submit_response.request.body,
                '\nResponse', submit_response.content)
    return submit_response.text 

def parse_result(result):
    """ Parse ANAT server response. Returns pd.DataFrame with edges or None if no edges were found
    """
    network_graph = result[0][0]
    edges = {}
    for x in network_graph:
        if x.tag == '{network}edges':
            [edges.setdefault(a.tag.replace('{network}', ''), [])
                    .append(a.text) for a in x]
    edges = pd.DataFrame(edges)
    try:
        subnetwork = (edges[['id1', 'id2']]
            .rename(columns={'id1' : 's', 'id2' : 't'}))
        return subnetwork
    except KeyError:
        return None


def network(sessionId, terminals, anchor=None, **kwargs):
    if anchor is None:
        submit_response = submit_job_no_anchor(sessionId, terminals, **kwargs)
    else:
        submit_response = submit_job(sessionId, anchor, terminals, **kwargs)
    got_result = False
    while not got_result:
        result = ET.fromstring(get_result(sessionId))
        got_result = len(result[0][0]) > 0
        time.sleep(0.5)
    return parse_result(result)


def remote_network(sessionId, network_undirected, terminals, anchor=None, **kwargs):
    if anchor is None:
        submit_response = submit_job_no_anchor(sessionId, network_undirected, terminals, **kwargs)
    else:
        submit_response = submit_job(sessionId, network_undirected, anchor, terminals, **kwargs)
    retries = 500 # more than half an hour
    got_result = False
    while not got_result and retries > 0:
        result = ET.fromstring(get_result(sessionId))
        got_result = len(result[0][0]) > 0
        time.sleep(5)
        retries = retries - 1
    if retries <= 0:
        print("Failed to obtain ANAT results. Exceeded retry limit of 500", flush=True)
        return None
    return parse_result(result)




import pandas as pd

def load(confidence, degree_threshold, db, ppi_network = None):
    """ read the network table from file
    filter for low confidence edges and high-degree nodes """
    ppi_network = db['ppi-network'] if ppi_network is None else ppi_network
    _anat = pd.read_table(ppi_network,
            names=['kin', 'sub', 'confidence', 'd'])
    anat = _anat[_anat['confidence'] > confidence]
    # Use degree cutoff to remove high-degree nodes
    _degk = anat.groupby('kin').apply(len)
    _degs = anat.groupby('sub').apply(len)
    _deg = pd.concat([_degk, _degs], axis=1).sum(axis=1)
    _low_deg = _deg[_deg < degree_threshold].index
    anat = anat[anat['kin'].isin(_low_deg) & anat['sub'].isin(_low_deg)].reset_index(drop=True)
    return anat

def to_directed(anat):
    """ duplicate undirected edges to form a directed network """
    found_in = ['found_in'] if 'found_in' in anat else []
    anat_directed = (pd.concat([anat, anat.rename(columns={'kin':'sub', 'sub':'kin'})])
            .reset_index(drop=True)
            [['kin', 'sub', 'confidence', 'd'] + found_in])
    anat_directed['d'] = 1
    return anat_directed

def add_unbiased_confidence(anat, evidence_path):
    """add large_scale evidence as confidence replacement to remove phosphositeplus bias
    
    >>> unbiased = add_unbiased_confidence(anat, 'db/anat/H_sapiens.txt')
    """
    evidence = pd.read_table(evidence_path)
    SOURCES = {'Y2H' : 'MI:0018'}
    evidence['found_in'] = evidence[list(SOURCES.values())].sum(axis=1)
    return pd.merge(anat, evidence, left_on=['kin', 'sub'],
            right_on=['interactor_a', 'interactor_b'])[['kin', 'sub', 'confidence', 'd', 'found_in']]





import pandas as pd
import numpy as np

from phos.util import fdrcorrection

def preprocess(exp, network, min_size):
    _kins = (pd.merge(exp, network, left_on='GeneID', right_on='sub')
            .groupby('kin')
            .apply(len))
    kins = _kins[_kins > min_size]
    filtered_network = network[network['kin'].isin(kins.index)]
    return kins, filtered_network

def run_one_sample_regular_weights(exp, network, beta=0):
    """
    weighted mean and standard deviation
    inspired by http://www.gnu.org/software/gsl/

    FIXME: remove beta
    """
    df = (pd.merge(exp, network, left_on='GeneID', right_on='sub')
            .rename(columns={'avg' : 'x', 'confidence' : 'w'}))
    if df.shape[0] == 0:
	    raise ValueError("None of the 'GeneID' could be matched to the network. Only human entrez GeneIDs are supported.")
    df['wx'] = df['w'] * df['x']
    df['w(x^2)'] = df['w'] * np.power(df['x'], 2)
    df['w^2'] = np.power(df['w'], 2)
    kin = df.groupby('kin')
    n = kin.apply(len)
    V1 = kin['w'].sum()
    V2 = kin['w^2'].sum()
    var = kin['w(x^2)'].sum() / (V1 - (V2 / V1))
    std = np.sqrt(var)
    mean = kin['wx'].sum() / V1
    t = mean / (np.power(std, beta) / np.sqrt(n))
    return t.to_frame(name='t')

def empiric(exp, network, min_size, permutations, side):
    num_neighbors, fnetwork = preprocess(exp, network, min_size)
    run = run_one_sample_regular_weights
    _exp = exp.reset_index(drop=True)
    emp = run(_exp, fnetwork)
    emp['pos'] = 0
    emp['neg'] = 0
    for _ in range(permutations):
        rnd = (_exp['avg']
                .reindex(np.random.permutation(_exp.index))
                .reset_index(drop=True))
        if type(rnd) is pd.Series:
            rnd = rnd.to_frame(name='avg')
        rnd['GeneID'] = _exp['GeneID']
        rnd_act = run(rnd, fnetwork)
        emp['pos'] = emp['pos'] + (emp['t'] < rnd_act['t'])
        emp['neg'] = emp['neg'] + (emp['t'] >= rnd_act['t'])
    emp['p_greater'] = emp['pos'] / permutations
    emp['p_lesser'] = emp['neg'] / permutations
    emp['p_twosided'] = emp[['p_greater', 'p_lesser']].min(axis=1)
    rej, q = fdrcorrection(emp['p_greater'].append(emp['p_lesser']))
    n = emp.shape[0]
    emp['rej_greater'] = rej[:n]
    emp['rej_lesser'] = rej[n:]
    emp['rej_twosided'] = rej[:n] | rej[n:]
    
    emp['q_greater'] = q[:n]
    emp['q_lesser'] = q[n:]
    emp['q_twosided'] = np.min([q[:n], q[n:]], axis=0)

    mask = emp['p_greater'] < emp['p_lesser']
    # continuity correction with 1/permutations to avoid log(0) = -Inf
    emp['score_empiric'] = ((mask * -np.log(emp['p_twosided'] + 1/permutations))
            + (~mask * np.log(emp['p_twosided'] + 1/permutations)))
    if side not in {'greater', 'twosided', 'lesser'}:
        raise ValueError('side parameters has to be one of "greater", "twosided", "less", was {}'.format(side))
    emp['Significant'] = emp['rej_{}'.format(side)]
    return (emp.drop(['pos', 'neg'], 1)
            .reset_index()
            .rename(columns={'kin' : 'GeneID'}))



import phos.data.read
import phos.data.network

def get_experiment(experiment_type):
    if experiment_type == 'silac':
        from phos.data.silac import experiment
        return experiment
    elif experiment_type == 'labelfree':
        from phos.data.labelfree import experiment
        return experiment
    else:
        raise NotImplementedError('experiment type {} unknown'.format(experiment_type))



""" taken from statsmodels """
import numpy as np

def _ecdf(x):
    nobs = len(x)
    return np.arange(1,nobs+1)/float(nobs)

def fdrcorrection(pvals, alpha=0.05):
    pvals = np.asarray(pvals)
    pvals_sortind = np.argsort(pvals)
    pvals_sorted = np.take(pvals, pvals_sortind)
    ecdffactor = _ecdf(pvals_sorted)
    reject = pvals_sorted <= ecdffactor*alpha
    if reject.any():
        rejectmax = max(np.nonzero(reject)[0])
        reject[:rejectmax] = True
    pvals_corrected_raw = pvals_sorted / ecdffactor
    pvals_corrected = np.minimum.accumulate(pvals_corrected_raw[::-1])[::-1]
    del pvals_corrected_raw
    pvals_corrected[pvals_corrected>1] = 1
    pvals_corrected_ = np.empty_like(pvals_corrected)
    pvals_corrected_[pvals_sortind] = pvals_corrected
    del pvals_corrected
    reject_ = np.empty_like(reject)
    reject_[pvals_sortind] = reject
    return reject_, pvals_corrected_



"""
tools for reading configuration info and data
"""
import os

import pandas as pd

import phos.data
import phos.util.io as io

def load(input, setting, experiment, uniprot_mapping, geneinfo, **kwargs):
    """
    read the dataset from file as specified

    >>> dataset, aux = config('egf.py')
    >>> exp = load(dataset, aux)

    :param dataset: dictionary
    :param aux: dictionary
    :returns exp: pandas.DataFrame
    """
    cache = experiment.get('cache', None)
    if cache is not None and os.path.isfile(cache):
        exp = pd.read_pickle(cache)
    else:
        df = io.read(**input)
        experiment_func = phos.data.get_experiment(setting)
        exp = experiment_func(df, uniprot_mapping=uniprot_mapping, geneinfo=geneinfo,
                **experiment)
        if cache is not None:
            exp.to_pickle(cache)
    return exp

def config(configfile):
    """
    read a configfile and return the dataframe
    
    >>> configs = config('egf.py')

    :param configfile: the path to the configuration file
    :returns dataset: the dictionary with the configu info
    """
    configs = {}
    with open(configfile) as f:
        exec(f.read(), configs)
    return configs



import pickle
import os

import pandas as pd

def filter_phospho(df):
    """ filter and format phosphosite extra tables """
    is_human = df['ORGANISM'] == 'human'
    ptms = df['MOD_RSD'].str.split('-')
    ptm_type = ptms.str.get(1)
    ptm_rsd = ptms.str.get(0)
    ptm_aa = ptm_rsd.str.get(0)
    ptm_pos = ptm_rsd.str.slice(1).astype(int)
    ptm_split = pd.concat([df['GENE_ID'], ptm_aa, ptm_pos], axis=1)
    phospho = ptm_split[(ptm_type == 'p') & is_human]
    phospho.columns = ['GeneID', 'Amino.Acid', 'Position']
    phospho = phospho.dropna()
    phospho['GeneID'] = phospho['GeneID'].astype(int)
    return phospho.drop_duplicates()

def _rename_amino_acid(df):
    """ 3 letter to single letter amino-acid code"""
    STY = {'Ser' : 'S', 'Thr' : 'T', 'Tyr' : 'Y'}
    amino_acid = STY.get(df['ORIGINAL'].iloc[0], float('NaN'))
    df['Amino.Acid'] = amino_acid
    return df

def _phosphosite_extra(filename, organism):
    """ parse phosphosite data """
    df = pd.read_csv(filename)
    if organism is not None:
        if hasattr(organism, '__iter__'):
            mask = df.ORGANISM.isin(organism)
        else:
            mask = df.ORGANISM == organism
        df.drop(df.index[~mask], inplace=True)
    return df

def regulatory_phosphosites(filename, organism=None, **kwds):
    return _phosphosite_extra(filename, organism)

def disease_associated_phosphosites(filename, organism=None, **kwds):
    return _phosphosite_extra(filename, organism)

def uniprot_humsavar(filename):
    """
    read amino-acid polymorphisms from
    http://uniprot.org/docs/humsavar
    """
    import re
    import io
    out = io.StringIO()
    out.write('\t'.join(['GENE_NAME', 'ACC', 'FTId', 'ORIGINAL',
        'POSITION', 'MUTATED', 'VARIANT', 'dbSNP', 'DISEASE']) + '\n')
    with open(filename) as f:
        s = re.compile('(\S+)\s+(\S+)\s+(\S+)\s+p\.([a-zA-Z]+)(\d+)([a-zA-Z]+)\s+(\S+)\s+(\S+)\s+(.*)') # one regex to rule them all
        for i,line in enumerate(f):
            if i < 30: # header
                continue
            if line == '\n': # reached end
                break
            out.write('\t'.join(s.match(line.strip()).groups()) + '\n')
    out.seek(0)
    return pd.read_table(out, na_values='-')

def read_files(db):
    """ return functional sites sources
    dis - disease associated sites from uniprot
    reg - regulatory sites from uniprot
    muts - SAPs
    """
    import phos.data.mapping as mapping
    dis = filter_phospho(disease_associated_phosphosites(db['phosphosite']['disease_associated_sites']))
    reg = filter_phospho(regulatory_phosphosites(db['phosphosite']['regulatory_sites']))
    _muts = (uniprot_humsavar(db['uniprot_humsavar']).dropna()
            .groupby('ORIGINAL').apply(_rename_amino_acid).dropna()
            .rename(columns={'POSITION': 'Position'}))
    muts = mapping.map_protein_groups(_muts, 'ACC')[['GeneID', 'Amino.Acid', 'Position']]
    return dis, reg, muts

def functional_sites(db):
    _cache = '.functional_sites.pkl'
    if os.path.isfile(_cache):
        with open(_cache, 'rb') as f:
            functional_sites = pickle.load(f)
    else:
        dis, reg, muts = read_files()
        functional_sites = (dis.append(reg).append(muts).drop_duplicates())
        with open(_cache, 'wb') as f:
            pickle.dump(functional_sites, f)



import pandas as pd
import numpy as np
import scipy
from sklearn import metrics
from sklearn import cross_validation as cv
from sklearn.linear_model import LogisticRegression


def logistic_regression(df, features, N=10):
    """ run logistic regression model
    
    :param df: pandas.DataFrame: inlcudes features and label column 
    :param features: list of features on which to train
    :param N: number of cross-validations
    :returns (all_probas, coef, mean_auc, classifier, fig):
        prediction probabilities, mean coefficients, mean AUC of the ROC,
        matplotlib figure of the ROC
    """
    X = df[features].values
    y = df['label'].values
    classifier = LogisticRegression(class_weight='balanced')
    kfold = cv.StratifiedKFold(y, n_folds=N)
    mean_tpr = 0.0
    mean_fpr = np.linspace(0, 1, 100)
    all_probas = np.zeros_like(y, dtype=float)
    all_coef = []
    for i, (train, test) in enumerate(kfold):
        probas = classifier.fit(X[train], y[train]).predict_proba(X[test])
        all_probas[test] = probas[:,1]
        all_coef.append(classifier.coef_[0])
        fpr, tpr, _ = metrics.roc_curve(y[test], probas[:, 1])
        mean_tpr += scipy.interp(mean_fpr, fpr, tpr)
        mean_tpr[0] = 0.0
    mean_tpr = mean_tpr / len(kfold)
    mean_tpr[-1] = 1.0
    mean_auc = metrics.auc(mean_fpr, mean_tpr)
    coef = pd.DataFrame(all_coef, columns=features).mean()
    return all_probas, coef, mean_fpr, mean_tpr, mean_auc, classifier



import os

import pandas as pd

import phos.algo.logistic_regression as clf
import phos.util.mapping as mapping

def predict_functional_sites(exp, scores, db, _functional_features = ['num_sites'], **kwargs):
    num_sites = get_sites_per_geneid(db['uniprot_mapping'],
            **db['phosphosite'])
    predicted_functional = (exp.merge(scores, on='GeneID')
            .merge(num_sites, on='GeneID'))
    known_functional = get_functional_sites(**db['phosphosite'])
    known_functional['functional'] = 1.0
    merged = pd.merge(predicted_functional, known_functional, how='left')
    predicted_functional['label'] = merged['functional'].notnull()
    features = ['avg'] + _functional_features + ['score_empiric']
    if merged.shape[0] > 10:
        res = clf.logistic_regression(predicted_functional, features)
        probas, coefs, mean_fpr, mean_tpr, mean_auc, classifier = res
    else:
        probas, coefs, mean_fpr, mean_tpr, mean_auc, classifier = (float('NaN'), [], float('NaN'), float('NaN'), float('NaN'), None)
    predicted_functional['proba'] = probas
    return predicted_functional, coefs, mean_fpr, mean_tpr, mean_auc, classifier

def get_sites_per_geneid(uniprot_mapping, all_sites, **kwargs):
    """ calculate how many known sites are on each geneid.
    based on all sites reported in phosphositeplus

    :param all_sites: path to the phosphorylation site dataset
    :param uniprot_mapping: path to uniprot mapping file
    :returns num_sites:
    """
    cache = all_sites + '.pkl'
    if os.path.isfile(cache):
        mapped_sites = pd.read_pickle(cache)
    else:
        sites = pd.read_table(all_sites, compression='gzip', skiprows=3)
        human_sites = sites[sites['ORGANISM'] == 'human']
        mapped_sites = mapping.map_protein_groups(human_sites, 'ACC_ID',
                'GeneID', uniprot_mapping)
        mapped_sites.to_pickle(cache)
    num_sites = (mapped_sites[['GeneID', 'MOD_RSD']]
            .drop_duplicates()
            .groupby('GeneID')
            .apply(len)
            .to_frame(name='num_sites')
            .reset_index())
    return num_sites

def get_functional_sites(regulatory_sites, disease_associated_sites, **kwargs):
    """ load known functional sites from phosphosite
    (i) regulatory sites (ii) disease-associated sites

    :param regulatory_sites: file path
    :param disease_associated_sites: file path
    """
    reg = _format_phosphosite_table(pd.read_csv(regulatory_sites))
    da = _format_phosphosite_table(pd.read_csv(disease_associated_sites))
    functional = pd.concat([reg, da]).drop_duplicates()
    return functional

# HELPER FUNCTIONS

def _format_phosphosite_table(df):
    """ filter and format phosphosite extra tables """
    is_human = df['ORGANISM'] == 'human'
    ptms = df['MOD_RSD'].str.split('-')
    ptm_type = ptms.str.get(1)
    ptm_rsd = ptms.str.get(0)
    ptm_aa = ptm_rsd.str.get(0)
    ptm_pos = ptm_rsd.str.slice(1).astype(int)
    ptm_split = pd.concat([df['GENE_ID'], ptm_aa, ptm_pos], axis=1)
    phospho = ptm_split[(ptm_type == 'p') & is_human]
    phospho.columns = ['GeneID', 'Amino.Acid', 'Position']
    phospho = phospho.dropna()
    phospho['GeneID'] = phospho['GeneID'].astype(int)
    return phospho.drop_duplicates()

def _rename_amino_acid(df):
    """ 3 letter to single letter amino-acid code"""
    STY = {'Ser' : 'S', 'Thr' : 'T', 'Tyr' : 'Y'}
    amino_acid = STY.get(df['ORIGINAL'].iloc[0], float('NaN'))
    df['Amino.Acid'] = amino_acid
    return df



""" utility functions for id mapping """
import os.path

import pandas as pd

def map_protein_groups(data, uniprot, geneid, uniprot_mapping, astype=int):
    """ turn protein group dataset into gene datset

    >>> genes = map_protein_groups(data, 'Uniprot ID', 'GeneID', 'idmapping.dat.gz')
    
    :param data: pandas.DataFrame to be mapped
    :param uniprot: string with the uniprot column name
    :param geneid: string specifying the target geneid column name,
                    has to correspond to idmapping.dat db name
    :param uniprot_mapping: string with the path to the uniprot id-mapping file
    :param astype: convert mapped ids to type
    :returns data: mapped"""
    uni2gid = dict(load_uniprot(uniprot_mapping, geneid).values)
    data[geneid] = data[uniprot].apply(_mapper, args=(uni2gid,))
    data = data.dropna(subset=[geneid])
    data = split_and_stack(data, geneid, ';')
    data[geneid] = data[geneid].astype(astype)
    return data

def _mapper(unis, uni2gid):
    """ maps uniprot to geneid and joins on ';' """
    if type(unis) is float:
        return uni2gid.get(unis, None)
    mapped = {uni2gid.get(u, None) for u in unis.split(';')}
    filtered = [m for m in mapped if m is not None]
    if len(filtered) == 0:
        return None
    return ';'.join(filtered)



def load_uniprot(filename, db='GeneID', force=False):
    """
    read unprot mapping from file. cache uniprot table to save time
    
    :param filename: gzipped uniprot mapping file path HUMAN_9606_idmapping.dat.gz
    :param force: force reparsing the file
    :returns uniprot: pandas.DataFrame mapping table
    """
    cache = '{}_{}_.pkl'.format(filename, db)
    if os.path.isfile(cache) and not force:
        uniprot = pd.read_pickle(cache)
    else: 
        uniprot = parse_uniprot_mapping(filename, db)
        uniprot.to_pickle(cache)
    return uniprot

def parse_uniprot_mapping(filename, db='GeneID'):
    """
    read uniprot mapping from file
    
    :param filename: gzipped uniprot mapping file path HUMAN_9606_idmapping.dat.gz
    """
    raw_data = pd.read_table(filename, compression='gzip',
            names=['Uniprot', 'db', 'dbid'])
    geneid_data = raw_data[raw_data['db'] == db]
    mapping = pd.pivot_table(geneid_data, 'dbid', index='Uniprot', columns='db',
            aggfunc=lambda x : ';'.join(x)).reset_index()
    return mapping

def split_and_stack(frame, colname, sep=None):
    """
    Converts a row of delimited values in a dataframe to multiple rows.
    The resulting table is useful for joins.

    >>> import pandas as pd 
    >>> df = pd.DataFrame.from_items([('col1', ['1 2 3']), ('col2', ['a'])])
    >>> splitted = split_and_stack(df, 'col1')
    >>> splitted.shape
    (3, 2)

    >>> splitted
      col2 col1
    0    a    1
    0    a    2
    0    a    3

    :param frame: the data frame to process
    :param colname: the name of the column containing the delimited data
    :param sep: an optional pattern for split(). if not specified split on whitespace
    """
    mask = frame[colname].isnull()

    masked_frame=frame[~mask]
    mapping = pd.DataFrame([{i: v for i,v in enumerate(x)} for x in
                             masked_frame[colname].str.split(sep).tolist()],
                             index=masked_frame.index).stack().reset_index(1,drop=True)
    mapping.name = colname # required for the merge

    return frame.drop(colname, axis=1).join(mapping)




import phos.util.mapping
import phos.util.io
import json
from collections import OrderedDict
from phos.util.fdr import fdrcorrection

def read_json(filename):
    with open(filename) as f:
        return json.load(f, object_pairs_hook=OrderedDict)

def deep_update(old, new):
    """ updates the second level of a dictionary

    >>> d = {'a' : {'b' : 1, 'c': 2}}
    >>> deep_update(d, {'a' : {'b' : 2}})
    >>> d
    {'a' : {'b' : 2, 'c' : 2}}
    """
    for k, v in new.items():
        old[k].update(v)

def num(string):
    """ convert string to int/float
    
    >>> type(num(1))
    int
    >>> type(num(1.5))
    float
    """
    try:
        return int(string)
    except ValueError:
        return float(string)

def uuid():
    import uuid
    return str(uuid.uuid4())



from setuptools import setup, find_packages

import os
HERE = os.path.dirname(__file__)
def read(fname):
    return open(os.path.join(HERE, fname)).read()

# creates version_string
exec(open(os.path.join(HERE, "phos", "version.py")).read())

setup(name='photon_ptm',
        version=version_string,
        description='PHOsphoproteomic dissecTiOn via Networks',
        long_description=read('README.md'),
        url='http://www.github.com/jdrudolph/photon',
        author='Jan Rudolph',
        author_email='jan.daniel.rudolph@gmail.com',
        license='MIT',
        packages=find_packages(),
        install_requires=[
            'numpy',
            'pandas>=0.21',
            'scipy',
            'matplotlib',
            'networkx',
            'perseuspy==0.3.5',
            'flask',
            'requests',
            'celery',
            'redis',
            'goenrich',
            'joblib',
            'sklearn',
            'pytest-runner'],
        tests_require=['pytest'])



""" reading and writing data """
import pandas as pd
import os

def read(path, format, NaN=None, data=None, query=None, cache=None, **kwargs):
    """ read dataset raw file
    >>> df = read(**dataset['file']) """
    if cache is not None:
        if os.path.isfile(cache):
            raw = pd.read_pickle(cache)
        else:
            raw = _read(path, format, NaN, data, query, **kwargs)
            raw.to_pickle(cache)
    else:
        raw = _read(path, format, NaN, data, query, **kwargs)
    return raw

def _read(path, format, NaN=None, data=None, query=None, **kwargs):
    if format == 'tsv':
        raw = pd.read_table(path, **kwargs)
    elif format == 'xlsx':
        raw = pd.read_excel(path, **kwargs)
    else:
        raise NotImplementedError('unknown data format', format)
    if NaN is not None:
        raw = raw.mask(raw == NaN)
    if data is not None:
        raw = raw.dropna(subset=data)
    if query is not None:
        raw = raw.query(query)
    return raw
